-- compute
#version 450 core

	// Sort 16 sortable elements
#define NUM_ELEMENTS 16

layout ( local_size_x = NUM_ELEMENTS / 2 ) in;

layout ( binding = 0 ) buffer SortData 
{
	uint value [];
};

	// Use local memory to minimize memory access cost.
shared uint local_value [NUM_ELEMENTS];

void	localCompareAndSwap ( uvec2 idx )
{
	if ( local_value [idx.x] > local_value [idx.y] ) 
	{
		uint tmp = local_value [idx.x];

		local_value [idx.x] = local_value [idx.y];
		local_value [idx.y] = tmp;
	}
}

void	flip ( uint h )
{
	uint	t       = gl_LocalInvocationID.x;
	uint	q       = ((2 * t) / h) * h;
	uint	t2      = t % (h / 2);
	uvec2	indices = q + uvec2 ( t2, h - t2 - 1 );

	localCompareAndSwap ( indices );
}

void	disperse ( uint h )
{
	uint	t       = gl_LocalInvocationID.x;
	uint	q       = ((2 * t) / h) * h;
	uint	t2      = t % (h / 2);
	uvec2	indices = q + uvec2 ( t2, t2 + (h / 2) );

	localCompareAndSwap ( indices );
}

void	main ()
{
	uint	t = gl_LocalInvocationID.x;
 	int		n = NUM_ELEMENTS;

		// each thread loads two elements to shared memory
	local_value [t*2    ] = value [t*2    ];
	local_value [t*2 + 1] = value [t*2 + 1];
 
	for ( uint h = 2; h <= n; h *= 2 )
	{
		barrier ();
		flip    ( h );

		for ( uint hh = h / 2; hh > 1; hh /= 2 ) 
		{
			barrier  ();
			disperse ( hh );
		}
	}

	barrier ();

		// write data back
	value [t*2    ] = local_value [t*2    ];
	value [t*2 + 1] = local_value [t*2 + 1];
}
