#include	<algorithm>
#include	"GlutWindow.h"
#include	"Program.h"
#include	"VertexArray.h"
#include	"VertexBuffer.h"
#include	"glUtilities.h"

#define	NUM_ELEMENTS	(1024*1024)

#define LocalBitonicMergeSort	0
#define LocalDisperse           1
#define BigFlip                 2
#define BigDisperse             3

class	SortWindow : public GlutWindow
{
	Program					program;
	VertexArray				vao;
	VertexBuffer			buf;
	std::vector<uint32_t>	array;
	const int				blockSize = 512;

public:
	SortWindow () : GlutWindow ( 100, 100, 512, 512, "Compute Sort" )
	{
			// fill array with random numbers
		array.resize ( NUM_ELEMENTS );

		for ( int i = 0; i < NUM_ELEMENTS; i++ )
			array [i] = rand ();

			// load sortf shader
		if ( !program.loadProgram ( "global-sort.glsl" ) )
			exit ( "Error building program: %s\n", program.getLog ().c_str () );	
		
			// create buffer and fill with data
		program.bind  ();
		vao.create    ();
		vao.bind      ();
		buf.create    ();
		buf.bindBase  ( GL_SHADER_STORAGE_BUFFER, 0 );
		buf.setData   ( NUM_ELEMENTS * sizeof ( uint32_t ), array.data (), GL_DYNAMIC_DRAW );

			// sort VBO data
		sort ( NUM_ELEMENTS );

		program.unbind    ();

			// get data from GPU back to CPU
		std::vector<uint32_t>	buffer ( NUM_ELEMENTS );

		glGetNamedBufferSubData ( buf.getId (), 0, NUM_ELEMENTS * sizeof ( uint32_t ), buffer.data () );

		std::sort ( array.begin (), array.end () );
/*
		printf ( "True sorted:\n\t" );

		for ( int i = 0; i < 32; i++ )
			printf ( "%d ", array [i] );

		printf ( "\nOur sorted:\n\t" );

		for ( int i = 0; i < 32; i++ )
			printf ( "%d ", buffer [i] );
*/
			// compare CPU and GPU sorted data
		printf ( "\nDiffs:\n" );

		for ( int i = 0; i < NUM_ELEMENTS; i++ )
			if ( buffer [i] != array [i] )
				printf ( "Diff aat index %d - %d vs %d\n", i, array [i], buffer [i] );
	}

	void redisplay () override
	{
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	}

	void	dispatch ( int n, int h, int algo )
	{
		const int blockCount = n / ( blockSize * 2 );

		program.setUniformInt ( "step",      h    );
		program.setUniformInt ( "algorithm", algo );

		glDispatchCompute ( blockCount, 1, 1 );
		glMemoryBarrier   ( GL_SHADER_STORAGE_BARRIER_BIT );
	}

	void	sort ( int n )
	{
		int	h = blockSize * 2;

		assert ( h <= n );
		assert ( h % 2 == 0 );

		dispatch ( n, h, LocalBitonicMergeSort );

		for ( h *= 2; h <= n; h *= 2 )
		{
			dispatch ( n, h, BigFlip );

			for ( int hh = h / 2; hh > 1; hh /= 2 )
			{
				if ( hh <= blockSize * 2 ) 
				{
					// can we fit all elements for a disperse operation into shared memory
					dispatch ( n, hh, LocalDisperse );
					break;
				} 
				else
					dispatch ( n, hh, BigDisperse );
			}
		}
	}

	void	sortSlow ( int n )
	{
		for ( int h = 2; h <= n; h *= 2 ) 
		{
			dispatch        ( n, h, BigFlip );

			for ( int hh = h / 2; hh > 1; hh /= 2 ) 
				dispatch        ( n, hh, BigDisperse );
		}
	}
};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv );
	
	SortWindow	win;
	
	return 0;
}
