#include	<algorithm>
#include	"GlutWindow.h"
#include	"Program.h"
#include	"VertexArray.h"
#include	"VertexBuffer.h"
#include	"glUtilities.h"

#define	NUM_ELEMENTS	16

class	SortWindow : public GlutWindow
{
	Program			program;
	VertexArray		vao;
	VertexBuffer	buf;
	std::vector<uint32_t>	array;

public:
	SortWindow () : GlutWindow ( 100, 100, 512, 512, "Compute Sort" )
	{
		array.resize ( NUM_ELEMENTS );

		for ( int i = 0; i < NUM_ELEMENTS; i++ )
			array [i] = rand ();

		printf ( "Not-sorted:\n\t" );
		
		for ( int i = 0; i < NUM_ELEMENTS; i++ )
			printf ( "%d ", array [i] );

		printf ( "\n" );

		if ( !program.loadProgram ( "local-sort.glsl" ) )
			exit ( "Error building program: %s\n", program.getLog ().c_str () );	
		
		program.bind  ();
		vao.create ();
		vao.bind    ();
		buf.create   ();
		buf.bindBase ( GL_SHADER_STORAGE_BUFFER, 0 );
		buf.setData  ( NUM_ELEMENTS * sizeof ( uint32_t ), array.data (), GL_DYNAMIC_DRAW );
		//buf.unbind   ();
		vao.unbind ();

		checkGlError ( "create buffers" );

		glDispatchCompute ( 1 /*NUM_ELEMENTS / 2*/, 1, 1 );

		checkGlError ("dispatch" );

		program.unbind    ();

		glMemoryBarrier   ( GL_SHADER_STORAGE_BARRIER_BIT );

		uint32_t	buffer [NUM_ELEMENTS];

		glGetNamedBufferSubData ( buf.getId (), 0, NUM_ELEMENTS * sizeof ( uint32_t ), buffer );

		checkGlError ("get buffer" );

		printf ( "Sorted:\n\t" );

		for ( int i = 0; i < NUM_ELEMENTS; i++ )
			printf ( "%d ", buffer [i] );

		printf ( "\n" );

		//buf.unmap ();

	}

	void redisplay () override
	{
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	}
};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv );
	
	SortWindow	win;
	
	return 0;	//GlutWindow::run ();
}
